/*  kbstick.c 
 *
 * KBStick v0.8 (c) 2000-03 David A. Fulgham
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

/***                                                                      ***
 *** This program is the work of a hobbyist. You could have done it, too. ***
 ***                                                                      ***
 *
 * The author of this program, David A. Fulgham, may be reached for comments
 * or bug reports by email at: daive@ditch.org
 */
 
/*
 * Many apologies, code cleanup & comments in next release or two. B)
 */ 


#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <sys/stat.h>  /* Included for open() */
#include <fcntl.h>     /* Included for open() */

#include <X11/Xlib.h>

#include <linux/joystick.h>
#include <sys/ioctl.h>
#include <signal.h>

#define MAX_AXES 12	/* These are actually max _directions_  */
			/* Change these if you want or  */
#define MAX_BUTTONS 12	/* need to add buttons/axes  	*/

const char* axisEnvVars[] = { "KBS_Axis_0_Plus",
                              "KBS_Axis_1_Plus",
                              "KBS_Axis_2_Plus",
                              "KBS_Axis_3_Plus",
                              "KBS_Axis_4_Plus",
                              "KBS_Axis_5_Plus",
                              "KBS_Axis_0_Minus",
                              "KBS_Axis_1_Minus",
                              "KBS_Axis_2_Minus",
                              "KBS_Axis_3_Minus",
                              "KBS_Axis_4_Minus",
                              "KBS_Axis_5_Minus" };

const char* buttonEnvVars[] = { "KBS_Button_0",
                                "KBS_Button_1",
                                "KBS_Button_2",
                                "KBS_Button_3",
                                "KBS_Button_4",
                                "KBS_Button_5",
                                "KBS_Button_6",
                                "KBS_Button_7",
                                "KBS_Button_8",
                                "KBS_Button_9",
                                "KBS_Button_10",
                                "KBS_Button_11" };
int axisKeyCodes[MAX_AXES];
int buttonKeyCodes[MAX_BUTTONS];

int quitButton = 0;
int quitButtonPresses = 0;

Display* ourXDisplay = 0;

const int delayForSend = 10;	/* Delay in mSec for sending key msgs */

int deadZone = 10;		/* Value an axis must move beyond before it is recognized */

/* char* lockFileName;  BAD */
char newLockFileName[256];



void help(const char *thisApp) {
  
  printf("\n  Usage: %s joy_device\n\n", thisApp);
}

char* cmdLine(int argc, char *argv[]) {

  char* joyDevName;
  struct stat fileStats;
  mode_t fileModes;
  
  if (argc < 2) {
    help(argv[0]);
    exit(-1);
  }
  
  stat(argv[1], &fileStats);
  fileModes = fileStats.st_mode;
  
  if(!S_ISCHR(fileModes)) {
    help(argv[0]);
    printf(" The filename provided as the second argument to\n");
    printf(" %s doesn't seem to be a joystick device.\n\n", argv[0]);
    exit(1);

  } else {
    joyDevName = argv[1];
  }
  
  return joyDevName;    
}

int initJoystick(const char* joyDevName) {

  int joystick;
  char axes = 2;
  char buttons = 2;
  
  if ((joystick = open(joyDevName, O_RDONLY)) < 0 ) {
    perror("Error opening joystick for reading.");
    XCloseDisplay(ourXDisplay);
    unlink(newLockFileName); 
    exit(1);
  }
  ioctl(joystick, JSIOCGAXES, &axes);
  ioctl(joystick, JSIOCGBUTTONS, &buttons);
  printf("This joystick has %d axes and %d buttons\n", axes, buttons);

  return joystick;
}

Display* getXDisplay(const char* appName) {

  int ourEvent, ourError, ourMajor, ourMinor;
  
  Display* ourDisplay = XOpenDisplay(0);
  
  if (! ourDisplay) {
    printf("\n ERROR: This program is made to be run in X Windows.\n\n");
/*     perror("Could not open X display"); */
    unlink(newLockFileName);
    exit(1);
  }
  if (! XTestQueryExtension(ourDisplay, &ourEvent, &ourError, &ourMajor, &ourMinor) ) {
    printf("%s: ", appName);
    printf("XServer %s doesn't support the Xtest extensions we need...\n", DisplayString(ourDisplay));
    XCloseDisplay(ourDisplay);
    exit(1);
  }
  printf("\nServer %s supports XTest (version %d.%d). Good.\n\n", DisplayString(ourDisplay), ourMajor, ourMinor);
  
  XSync( ourDisplay, True);
  
  return ourDisplay;
}
  
void assignCodes() {

  int ourCounter;
  int codesAssigned = 0;
  char* result;
  
  result = (char*)getenv("KBS_Quit_Button");
  
  if(!result) {
    quitButton = 0;
    printf("\nWARNING: A quit button was not assigned.\n");
    printf(" You may set a quit button (quit KeyCode actually) by assigning\n");
    printf(" a value to the environment variable KBS_Quit_Button.\n");
    printf(" This is most easily done in the same place & after you set the\n");
    printf(" environment variables for your KeyCodes, i.e.:\n\n");
    printf("KBS_Button_5=108               # Set Button 5 to send KP_Enter\n");
    printf("KBS_Quit_Button=$KBS_Button_5  # Set Button 5 as the Quit button\n\n");
  }
  else {
    quitButton = atoi(result);
  }

  for (ourCounter= 0; ourCounter < MAX_AXES; ++ourCounter) {
/*  printf("%s value = %d\n", axisEnvVars[ourCounter], axisKeyCodes[ourCounter]); */
    result = (char*)getenv(axisEnvVars[ourCounter]);
/*    printf("Result of getting env value for %s is %s\n\n", axisEnvVars[ourCounter], result); */
    if (!result)
      axisKeyCodes[ourCounter] = 0;
    else {
      axisKeyCodes[ourCounter] = atoi(result);
      ++codesAssigned;
      }
/*    printf("%s value = %d\n", axisEnvVars[ourCounter], axisKeyCodes[ourCounter]); */    
  }
  for (ourCounter= 0; ourCounter < MAX_BUTTONS; ++ourCounter) {
/*    printf("%s value = %d\n", buttonEnvVars[ourCounter], buttonKeyCodes[ourCounter]); */
    result = (char*)getenv(buttonEnvVars[ourCounter]);
/*    printf("Result of getting env value for %s is %s\n", buttonEnvVars[ourCounter], result); */
    if (!result)
      buttonKeyCodes[ourCounter] = 0;
    else {
      buttonKeyCodes[ourCounter] = atoi(result);
      if(quitButton && buttonKeyCodes[ourCounter] == quitButton) 
        printf("\n%s is the quit button. Press 3 times to exit.\n\n", buttonEnvVars[ourCounter]);
      ++codesAssigned;
      }
/*    printf("%s value = %d\n", buttonEnvVars[ourCounter], buttonKeyCodes[ourCounter]); */

  }
  
  if (!codesAssigned) {
    printf("\nERROR: No KBS_ environment variables have been set.\n");
    printf(" This program reads the codes to assign to your joystick axes\n");
    printf(" and buttons from environment variables in the form of:\n\n");
    printf("KBS_Axis_0_Minus=[X KeyCode value]  #  Code sent when the stick goes\n");
    printf("                                    #   in the - X-Axis direction.\n\n");
    printf("KBS_Button_0=[X KeyCode value]  #  Code sent when stick button #0\n");
    printf("                                #  (i.e. the first joystick button)\n");
    printf("                                #  is pressed.\n");
    XCloseDisplay(ourXDisplay);
    unlink(newLockFileName);
    exit(1);
  }
/*  printf("\n\nQuit button KeyCode is: %s\n\n", result); */

  result = (char*)getenv("KBS_Dead_Zone");
  if(result) {
    deadZone = atoi(result);
    printf("  A Dead Zone of %d around center has \n", deadZone);
    printf("  been applied to *all* your joystick axes.\n\n");
    printf("  If you don't want this unset the \n");
    printf("  KBS_Dead_Zone environment varible.\n\n");
  }
  return;
}


void doXKey(Display* ourDisplay, const int keyToSend, int press) {

  if (press) 
    XTestFakeKeyEvent(ourDisplay, (unsigned int)keyToSend, True, delayForSend);
  else
    XTestFakeKeyEvent(ourDisplay, (unsigned int)keyToSend, False, delayForSend);
}



void readLoop(int jstick, char* appName, Display* ourDisplay) {

  int buttonDown;
  struct js_event joyReading;
  int axisMinusMask = 0;
  int axisPlusMask = 0;
  int quitButtonPressed = 0;
  
  while(!quitButtonPressed) {
  
    
    if( read(jstick, &joyReading, sizeof(struct js_event)) != sizeof(struct js_event)) {
      printf("%s: ", appName);
      perror("Error when trying to read from joystick.\n");
      if(ourDisplay)
        XCloseDisplay(ourDisplay);
      exit(1);
    }
    
    switch(joyReading.value) {
    
      case 0:
        if((joyReading.type) == JS_EVENT_BUTTON) {
          if(buttonKeyCodes[joyReading.number])
            doXKey(ourDisplay, buttonKeyCodes[joyReading.number], 0);
          break;
        }
        else if ((joyReading.type) == JS_EVENT_AXIS) {		/* An axis was released */
          if ((1 << joyReading.number) & axisMinusMask) {	/* if it's coming from - dir */
            if (axisKeyCodes[joyReading.number+(MAX_AXES/2)]){	/* If the value is set to something */
              doXKey(ourDisplay, 				/* Send the axis - KeyCode */
		axisKeyCodes[joyReading.number+(MAX_AXES/2)],
		 0);
            }
            axisMinusMask ^= (1 << joyReading.number);	     /* reset the - axis mask */
          /*  break;					      and break out */
          }
          if ((1 << joyReading.number) & axisPlusMask) {        /* Axis is coming from + dir */
            if (axisKeyCodes[joyReading.number]) {		/* If the value is set to something */
              doXKey(ourDisplay, 				/* Send the axis + KeyCode */
		axisKeyCodes[joyReading.number], 0);
            }
            axisPlusMask ^= (1 << joyReading.number);	  /* And reset the + axis mask */
          /*  break; */
          }
        }
        break;
        
      case 1:
        if((joyReading.type) == JS_EVENT_BUTTON) {
          if(buttonKeyCodes[joyReading.number]) {
            if (buttonKeyCodes[joyReading.number] == quitButton)
              ++quitButtonPresses;
            else
              quitButtonPresses = 0;
            if(quitButtonPresses > 2)
              quitButtonPressed = 1;
            doXKey(ourDisplay, buttonKeyCodes[joyReading.number], 1);
          }
          break;
        }          

          
        break;
        
      default:
          
        if ((joyReading.type) == JS_EVENT_AXIS) {
          if(joyReading.value > deadZone) {
            if((1 << joyReading.number) & axisPlusMask)	   /* If we already sent this KeyCode as a press... */
              break;					   /* Break out.  */
            if(axisKeyCodes[joyReading.number]) 
              doXKey(ourDisplay, axisKeyCodes[joyReading.number], 1);
            axisPlusMask |= (1 << joyReading.number);
            /* break; */
          }
          
          if(joyReading.value < -deadZone) {
            if((1 << joyReading.number) & axisMinusMask)	   /* If we already sent this KeyCode as a press...*/
              break;					        /* Break out  */
            if (axisKeyCodes[joyReading.number+(MAX_AXES/2)])
              doXKey(ourDisplay, axisKeyCodes[joyReading.number+(MAX_AXES/2)], 1);
            axisMinusMask |= (1 << joyReading.number);
            /* break; */
          }
        }
        break;
    }
  XFlush(ourDisplay);
  }
}      

void cleanup(const int signal) {
  XCloseDisplay(ourXDisplay);
  unlink(newLockFileName);
  printf("\nExiting on signal.\n");
  exit(0);
}

int main(int argc, char *argv[]) {

  char* joystickName;
/*  Display* ourXDisplay; */
  int joystick;
  char* prevRunning;
  int lockFile = 0;
  char* oldLockFileName;
  
  printf("\nKBStick v0.8  GPL (c) 2000-03 by David A. Fulgham\n");

  joystickName = cmdLine(argc, argv);
  
  oldLockFileName = (char*)getenv("HOME");
  if(!oldLockFileName)
    oldLockFileName = (char*)getenv("LOGDIR");
  if(!oldLockFileName) {
    printf("  Error: Neither HOME nor LOGDIR environment variables are set.\n");
    printf("  Please set one or the other to your home directory.\n");
    exit(1);
  }
  strcpy(newLockFileName, oldLockFileName);
  strcat(newLockFileName, "/.kbstick.lock");
  lockFile = open(newLockFileName ,O_RDONLY|O_CREAT|O_EXCL);
  if( lockFile < 0) {
    printf("\nA previous incarnation of %s is already running.\n", argv[0]);
    printf(" ( If you discover this was not the case \n");
    printf("   remove the file %s )\n", newLockFileName);
    exit(1);
  }
  else 
    close(lockFile);
    
  ourXDisplay = getXDisplay(argv[0]);

  signal(SIGINT, cleanup);
  signal(SIGTERM, cleanup);
  
  joystick = initJoystick(joystickName);
  
  assignCodes();
  
  readLoop(joystick, argv[0], ourXDisplay);
  
  close(joystick);
  XCloseDisplay(ourXDisplay);
  unlink(newLockFileName);
  
  printf("\nExiting normally.\n\n");
  exit(0);
}




  
